use strict;
use threads;
use threads::shared;
use Test::More tests => 39;
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::DBISQLite;

my $DB_FILE_NAME = 'DBISQLite.t.db';

## no critic qw(ProhibitMagicNumbers)
sub Main {
    my $dbi = DE_EPAGES::Database::API::DBISQLite->new( "dbi:SQLite:dbname=$DB_FILE_NAME" );

    ok( Init( $dbi ), 'Init' );
    ok( $dbi->existsTable( 'dbitest' ), 'existsTable' );
    ok( !$dbi->existsTable( 'notable' ), 'not existsTableColumn' );
    ok( $dbi->existsTableColumn( 'dbitest', 'name' ), 'existsTableColumn' );
    ok( !$dbi->existsTableColumn( 'dbitest', 'nocolumn' ), 'not existsTableColumn' );

    $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [10, 'Name1'] );
    my $Name = $dbi->execute( 'SELECT name FROM dbitest WHERE id=?', [10] )->[0]->[0];
    ok( $Name eq 'Name1', 'basic INSERT and SELECT' );

    $dbi->do( 'UPDATE dbitest SET name=? WHERE id=?', ['Name2', 10] );
    $Name = $dbi->execute( 'SELECT name FROM dbitest WHERE id=?', [10] )->[0]->[0];
    ok( $Name eq 'Name2', 'simple update' );

    # transaction with rollback
    my $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
    is( $Count, 1, 'count before transaction' );

    eval {
        $dbi->transaction( sub {
            $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [20, 'Name2'] );

            $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
            is( $Count, 2, 'count within transaction' );

            die 'fake error';
        } );
    };
    ok( $@ =~ /fake error/, 'transaction failed' );

    $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
    is( $Count, 1 , 'count after rolled back transaction' );

    # nested transaction with error in outer transaction
    eval {
        $dbi->transaction( sub {
            $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [20, 'Name2'], ['Integer', 'String'] );

            $dbi->transaction( sub {
                $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [30, 'Name3'], ['Integer', 'String'] );

                $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
                ok( 3 == $Count, "count within transaction (3 == $Count)" );
            } );

            die 'fake error';
        } );
    };
    ok( $@ =~ /fake error/, 'transaction failed' );

    $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
    ok( 1 == $Count, "count after transaction (1 == $Count)" );

    # transaction log part 1/3
    ok( !$dbi->isTransactionLogEnabled, 'isTransactionLogEnabled is false by default' );
    $dbi->enableTransactionLog(1);
    ok( $dbi->isTransactionLogEnabled, 'enableTransactionLog (1)' );

    # nested transaction with error in inner transaction
    eval {
        $dbi->transaction( sub {
            $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [20, 'Name2'], ['Integer', 'String'] );

            $dbi->transaction( sub {
                $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [30, 'Name3'], ['Integer', 'String'] );

                $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
                ok( 3 == $Count, "count within transaction (3 == $Count)" );

                # transaction log part 2/3
                ok( 2 == scalar @{$dbi->getTransactions}, 'transaction log constains 2 INSERT statements' );
                ok( $dbi->getTransactions->[1] eq 'INSERT INTO dbitest (id, name) VALUES (30, \'Name3\')', 'getTransactions' );

                die 'fake error';
            } );
        } );
    };
    ok( $@ =~ /fake error/, 'transaction failed' );

    # transaction log part 3/3
    ok( $dbi->isTransactionLogEnabled, 'transaction log is still enabled' );
    ok( 0 == scalar @{$dbi->getTransactions}, 'transaction log is empty because of rollback' );
    $dbi->enableTransactionLog(0);
    ok( !$dbi->isTransactionLogEnabled, 'enableTransactionLog(0)' );

    $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
    is( $Count, 1, 'count after transaction' );

    # date and time
    my $Now = $dbi->currentDateTime;
    is( $Now->time_zone_long_name, 'floating', 'currentDateTime has time zone "floating"' );
    ok( $Now->year > 2003, 'currentDateTime' );

    my $dt = $dbi->execute( 'SELECT dt FROM dbitest WHERE id=?', [10], ['DateTime'] )->[0]->[0];
    ok( !defined $dt, 'select NULL value' );

    $dbi->do( 'UPDATE dbitest SET dt=? WHERE id=?', [$Now, 10], ['DateTime', 'Integer'] );
    $dt = $dbi->execute( 'SELECT dt FROM dbitest WHERE id=?', [10], ['DateTime'] )->[0]->[0];
    is( $dt->hms, $Now->hms, 'equal time' );
    is( $dt->ymd, $Now->ymd, 'equal date' );
    is( $dt, $Now, 'update DATETIME value' );

    $dbi->do( 'DELETE FROM dbitest WHERE id=?', [10], ['Integer']  );
    my $aRows = $dbi->execute( 'SELECT name FROM dbitest WHERE id=?', [10] );
    is( scalar @$aRows, 0, 'deleted row' );

    # test for 'database is locked' (EPG-12856)
    $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
    is( $Count, 0, 'before test \'database is locked\'' );

    my $thread = async {
        my $dbi2 = DE_EPAGES::Database::API::DBISQLite->new( "dbi:SQLite:dbname=$DB_FILE_NAME" );
        $dbi2->transaction( sub {
            $dbi2->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [41, 'Name3'] );
            $dbi2->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [42, 'Name3'] );
            sleep 2;
        } );
    };

    eval {
        sleep 1;
        $dbi->transaction( sub {
            $dbi->execute( 'SELECT id, name FROM dbitest' );
            $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [51, 'Name4'] );
            $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [52, 'Name5'] );
            $dbi->do( 'INSERT INTO dbitest (id, name) VALUES (?, ?)', [53, 'Name6'] );
        } );
    };
    is( $@, '', 'no error' );
    $thread->join();

    $Count = $dbi->execute( 'SELECT COUNT(*) FROM dbitest', [] )->[0]->[0];
    is( $Count, 5, 'after test \'database is locked\'' );

    # DDL functions
    ok( !$dbi->existsTableColumn('dbitest', 'testcolumn'), 'before adding testcolumn');
    $dbi->addTableColumn('dbitest', 'testcolumn varchar(30) NULL');
    ok( $dbi->existsTableColumn('dbitest', 'testcolumn'), 'testcolumn added');

    ok( !$dbi->existsTableColumn('dbitest', 'isactive'), 'before adding column \'isactive\'');
    $dbi->addTableColumn('dbitest', 'isactive bit default 0 not null');
    ok( $dbi->existsTableColumn('dbitest', 'isactive'), 'isactive added');

    ok( Cleanup( $dbi ), 'Cleanup' );
    $dbi->disconnect;
    unlink $DB_FILE_NAME;
}

sub Init {
    my ( $dbi ) = @_;
    Cleanup( $dbi );

    $dbi->do( 'CREATE TABLE dbitest( id INTEGER, name VARCHAR(50) NULL, dt DATETIME NULL )' );

    return $dbi->existsTable( 'dbitest' );
}

sub Cleanup {
    my ( $dbi ) = @_;

    $dbi->do( 'DROP TABLE dbitest' ) if $dbi->existsTable( 'dbitest' );
    return 1;
}

RunScript(Sub=>\&Main, Debug=>1);
