use strict;
use Test::More tests => 70;
use DE_EPAGES::Database::API::Connection qw( ConnectTestSite DisconnectSite GetTestSite RunOnStore );
use DE_EPAGES::Object::API::Factory qw( LoadObjectByPath ExistsObjectByPath LoadRootObject );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::File qw (GetFileContent GetFileSize ExistsFile DeleteDirectory
                                   ExistsDirectory CreateDirectory );
use DE_EPAGES::Core::API::Error qw (ResetError ExistsError GetError);
use DE_EPAGES::Core::API::Image qw (GetImageInfo);

my $ROOTPATH = '/Test';

sub Init {
    Cleanup();
    my $Folder = LoadRootObject()->insertChild( 'Object', { Alias => 'Test' });
    my $Shop = $Folder->insertChild( 'Shop', { Alias => 'TestShop1' } );

    # set feature size for test:
    $Shop->featureMaxValue( 'GallerySize', 500000 );

    return 1;
}

sub Cleanup {
    LoadObjectByPath($ROOTPATH)->delete if ExistsObjectByPath($ROOTPATH);
    return 1;
}

#========================================================================================
# function     _testDirectoryMethods
# state        private
#----------------------------------------------------------------------------------------
# description  Runs various test for creating, renaming and deleting directories using
#               MediaGallery's methods.
#----------------------------------------------------------------------------------------
# input        $MediaGallery | object | MediaGallery child object of the shop.
#========================================================================================
sub _testDirectoryMethods {
    my ($MediaGallery) = @_;

    cmp_ok($MediaGallery->createDirectory('NewDir', 'MyFiles/SomeDir/'), 'eq', 'NewDir', "create dir");
    cmp_ok($MediaGallery->existsDirectory('MyFiles/SomeDir/NewDir'), 'eq', 1, "dir exists");
    cmp_ok($MediaGallery->renameDirectory('MyFiles/SomeDir/NewDir', 'Dir/With/GreatNewName'), 'eq', 'DirWithGreatNewName', "rename dir");
    cmp_ok($MediaGallery->existsDirectory('MyFiles/SomeDir/DirWithGreatNewName'), 'eq', 1, "dir exists");
    $MediaGallery->deleteDirectory('MyFiles/SomeDir/DirWithGreatNewName');
    cmp_ok($MediaGallery->existsDirectory('MyFiles/SomeDir/DirWithGreatNewName'), 'eq', 0, "after delete, dir is gone");

    cmp_ok($MediaGallery->createDirectory('d/r/eiAUOA/Sd1zz\\zz23', ''), 'eq', 'dreiAUOASd1zz_92_zz23', "many slashes");
    $MediaGallery->deleteDirectory('dreiAUOASd1zz_92_zz23');
    cmp_ok($MediaGallery->existsDirectory('dreiAUOASd1zz_92_zz23'), 'eq', 0, "after delete, dir is gone");

    cmp_ok($MediaGallery->createDirectory('NewDir', 'MediaGallery/MyFiles/SomeDir/'), 'eq', 'NewDir', "create dir w. MediaGallery");
    cmp_ok($MediaGallery->existsDirectory('MediaGallery/MyFiles/SomeDir/NewDir'), 'eq', 1, "dir exists");
    cmp_ok($MediaGallery->renameDirectory('MediaGallery/MyFiles/SomeDir/NewDir', 'MediaGallery/Dir/With/GreatNewName'), 'eq', 'MediaGalleryDirWithGreatNewName', "rename dir");
    cmp_ok($MediaGallery->existsDirectory('MediaGallery/MyFiles/SomeDir/MediaGalleryDirWithGreatNewName'), 'eq', 1, "dir exists");
    $MediaGallery->deleteDirectory('MediaGallery/MyFiles/SomeDir/MediaGalleryDirWithGreatNewName');
    cmp_ok($MediaGallery->existsDirectory('MediaGallery/MyFiles/SomeDir/MediaGalleryDirWithGreatNewName'), 'eq', 0, "after delete, dir is gone");

    cmp_ok($MediaGallery->createDirectory('Eins/Zwei/Drei', '', {'usebasedirectory' => 1}),
        'eq', 'EinsZweiDrei', "create dir using flags");
    $MediaGallery->deleteDirectory('EinsZweiDrei');
    cmp_ok($MediaGallery->existsDirectory('EinsZweiDrei'), 'eq', 0, "after delete, dir is gone");

    cmp_ok($MediaGallery->createDirectory('Eins/Zwei/Drei', undef, {'usebasedirectory' => 0}),
        'eq', 'Eins/Zwei/Drei', "create dir using flags");
    $MediaGallery->deleteDirectory('Eins');
    cmp_ok($MediaGallery->existsDirectory('Eins'), 'eq', 0, "delete method deletes recursively");

}

#========================================================================================
# function     _testRemainingMethods
# state        private
#----------------------------------------------------------------------------------------
# description  Runs various test for featues, strings and path conversions.
#----------------------------------------------------------------------------------------
# input        $Shop | object | Object of the shop.
# input        $MediaGallery | object | MediaGallery child object of the shop.
#========================================================================================
sub _testRemainingMethods {
    my ($Shop, $MediaGallery) = @_;

    cmp_ok($MediaGallery->relativePath($ENV{'EPAGES_WEBROOT'}.'/'.GetTestSite()."/Test/TestShop1/MediaGallery/SomeDir/DirWithGreatNewName/file.txt"),
        'eq', "SomeDir/DirWithGreatNewName/file.txt", "relativePath (MG's public path is: ".$MediaGallery->get("PublicPath").")");

    my $CorrectAbsPath = $ENV{'EPAGES_WEBROOT'}.'/'.GetTestSite()."/Test/TestShop1/MediaGallery/SomeDir/DirWithGreatNewName/file.txt";
    cmp_ok($MediaGallery->absolutePath("SomeDir/DirWithGreatNewName/file.txt"), 'eq',
        $CorrectAbsPath, "absolutePath");

    cmp_ok($MediaGallery->previewPath('file.txt', 'Thumbnail'), 'eq', '/file,Thumbnail.txt', "previewPath thumbnail");
    cmp_ok($MediaGallery->previewPath('SomeDir/DirWithGreatNewName/file.txt', 'mgt'), 'eq',
        'SomeDir/DirWithGreatNewName/file,mgt.txt', "previewPath mgt");

    # feature stuff

    # 1. set feature to 200/10KByte
    $Shop->featureCurrentValue('GallerySize', 200);
    $Shop->featureMaxValue('GallerySize', 10);
    # testGallerySize should throw an error
    eval {
        $MediaGallery->testGallerySize();
    };
    if(ExistsError()) {
        my $Error = GetError();
        ok($Error->code eq 'FeatureLimitExceeded', "testGallerySize: size exceeded");
    }
    # 2. modify max value (5MB)
    $Shop->featureMaxValue('GallerySize', 5000);
    # testGallerySize should now tell it's okay
    $MediaGallery->testGallerySize();
    ok(!ExistsError(), "testGallerySize: size increased, no more error");

    my $SizeBefore = $Shop->featureCurrentValue('GallerySize');
    # input is in Bytes!
    $MediaGallery->increaseGallerySize(157286400);
    # increaseGallerySize converts to KBytes!
    cmp_ok($Shop->featureCurrentValue('GallerySize'), 'eq', $SizeBefore+153600,
        "sucessfully increased GallerySize by 150KBytes");
}

#========================================================================================
# function     _testFileMethods
# state        private
#----------------------------------------------------------------------------------------
# description  Runs test for file operations (insert, delete, rename, move, copy, etc.)
#----------------------------------------------------------------------------------------
# input        $MediaGallery | object | MediaGallery child object of the shop.
#========================================================================================
sub _testFileMethods {
    my ($Shop, $MediaGallery) = @_;
    my $MGPublicPath = $MediaGallery->get('PublicPath');

    # insertFile()
    my $ContentString = "test file";
    cmp_ok($MediaGallery->insertFile('MyFiles/SomeDir/test.txt', \$ContentString), 'eq',
        'MyFiles/SomeDir/test.txt', "insertfile, new text file");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeDir/test.txt'), "file created successfully");
    my $FileContent = ${GetFileContent($MGPublicPath.'/MyFiles/SomeDir/test.txt')};
    cmp_ok($FileContent, 'eq', $ContentString, "created file's content ok");

    # this overwrites the file
    $ContentString = "overwritten content";
    cmp_ok($MediaGallery->insertFile('MyFiles/SomeDir/test.txt', \$ContentString, {overwrite=>1}), 'eq',
        'MyFiles/SomeDir/test.txt', "insertfile, overwrite text file");
    $FileContent = ${GetFileContent($MGPublicPath.'/MyFiles/SomeDir/test.txt')};
    cmp_ok($FileContent, 'eq', $ContentString, "overwritten file's content ok");

    # but this should not overwrite
    my $IgnoredContentString = "ignored string";
    cmp_ok($MediaGallery->insertFile('MyFiles/SomeDir/test.txt', \$IgnoredContentString), 'eq',
        'MyFiles/SomeDir/test1.txt', "insertfile, do not overwrite");
    $FileContent = ${GetFileContent($MGPublicPath.'/MyFiles/SomeDir/test.txt')};
    cmp_ok($FileContent, 'eq', $ContentString, "file's content still old");

    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeDir/test1.txt'), "created file with similar name");
    $FileContent = ${GetFileContent($MGPublicPath.'/MyFiles/SomeDir/test1.txt')};
    cmp_ok($FileContent, 'eq', $IgnoredContentString, "file's content ok");

    # existsFile()
    $ContentString = "exists test file";
    $MediaGallery->insertFile('MyFiles/SomeDir/existstest.txt', \$ContentString);
    ok($MediaGallery->existsFile('MyFiles/SomeDir/existstest.txt'), "exists file return true for existing file");
    ok(!$MediaGallery->existsFile('MyFiles/SomeDir/existstest09128309123801923.txt'), "exists file return false for non-existing file");

    # moveFile()
    $ContentString = "movefile test file";
    DeleteDirectory($MGPublicPath.'/MyFiles/SomeOtherDir/',{ recursive => 1 }) if ExistsDirectory($MGPublicPath.'/MyFiles/SomeOtherDir/');
    $MediaGallery->insertFile('MyFiles/SomeDir/movetest1.txt', \$ContentString);
    # this should throw an error as the target directory does not exists
    eval {
        $MediaGallery->moveFile('MyFiles/SomeDir/movetest1.txt', 'MyFiles/SomeOtherDir');
    };
    if( ExistsError() ) {
        my $Error = GetError();
        cmp_ok($Error->code, 'eq', "FILE_MOVE_FAILED", "move to non-existing dir throws error");
        ResetError();
    }
    # now should work
    CreateDirectory($MGPublicPath.'/MyFiles/SomeOtherDir', { 'recursive' => 1 });
    $MediaGallery->moveFile('MyFiles/SomeDir/movetest1.txt', 'MyFiles/SomeOtherDir');
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeDir/movetest1.txt'), "moveFile: old file is gone");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeOtherDir/movetest1.txt'), "moveFile: file now in new dir");

    # renameFile()
    $ContentString = "rename test file";
    CreateDirectory($MGPublicPath.'/MyFiles/SomeRenameDir', { 'recursive' => 1 });
    $MediaGallery->insertFile('MyFiles/SomeRenameDir/renametest.txt', \$ContentString);
    $MediaGallery->renameFile('MyFiles/SomeRenameDir/renametest.txt', 'othername123.txt');
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeRenameDir/renametest.txt'), "renameFile: old file is gone");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeRenameDir/othername123.txt'), "renameFile: file now with new name");

    # deleteFile()
    $ContentString = "delete test file";
    $MediaGallery->insertFile('MyFiles/SomeDir/deletetest1.txt', \$ContentString);
    $MediaGallery->deleteFile('MyFiles/SomeDir/deletetest1.txt');
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeDir/deletetest1.txt'), "deleteFile: file is gone");
    # rm directory as well
    CreateDirectory($MGPublicPath.'/MyFiles/SomeDeleteDir', { 'recursive' => 1 });
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/deletetest1.txt', \$ContentString);
    $MediaGallery->deleteFile('MyFiles/SomeDeleteDir/deletetest1.txt', 1);
    ok(!ExistsDirectory($MGPublicPath.'/MyFiles/SomeDeleteDir'), "deleteFile: dir is gone as well");

    # files with same name but different extensions (EPG-10961)
    $FileContent = ${GetFileContent('test.jpg', { binary => 1 })};
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/deletetest2.jpg', \$FileContent);
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/deletetest2,mgt.jpg', \$FileContent);
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/deletetest2.gif', \$FileContent);
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/deletetest2,mgt.gif', \$FileContent);
    $MediaGallery->deleteFile('MyFiles/SomeDeleteDir/deletetest2.gif');
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/deletetest2.gif'), "deleteFile: file with deleted");
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/deletetest2,mgt.gif'), "deleteFile: file thumbnail (mgt) deleted");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/deletetest2.jpg'), "deleteFile: file with same extension survived");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/deletetest2,mgt.jpg'), "deleteFile: file with same extension survived (mgt)");
    # files with "nearly" the same name (EPG-13475)
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/lang.jpg', \$FileContent);
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/lang,mgt.jpg', \$FileContent);
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/langauch.jpg', \$FileContent);
    $MediaGallery->insertFile('MyFiles/SomeDeleteDir/langauch,mgt.jpg', \$FileContent);
    $MediaGallery->deleteFile('MyFiles/SomeDeleteDir/lang.jpg');
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/lang.jpg'), "deleteFile: file with deleted");
    ok(!ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/lang,mgt.jpg'), "deleteFile: file thumbnail (mgt) deleted");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/langauch.jpg'), "deleteFile: file with almost same name survived");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeDeleteDir/langauch,mgt.jpg'), "deleteFile: file with almost same name survived (mgt)");

    # gallery2attributeFileName()
    cmp_ok($MediaGallery->gallery2attributeFileName('MyFiles/SomeDir/any.doc'), 'eq',
        'MediaGallery/MyFiles/SomeDir/any.doc', "gallery2attributeFileName");

    # attribute2galleryFileName()
    cmp_ok($MediaGallery->attribute2galleryFileName('MediaGallery/MyFiles/SomeDir/any.doc'), 'eq',
        'MyFiles/SomeDir/any.doc', "attribute2galleryFileName");

    # addUploadedFiles()
    my $TestDBImageName1 = 'animals_4237.jpg';
    my $TestDBImageName2 = 'animals_4249.jpg';
    my $TestDBImage1 = $ENV{'EPAGES_WEBROOT'}.'/ImageDB/original/EP_Images/'.$TestDBImageName1;
    my $TestDBImage2 = $ENV{'EPAGES_WEBROOT'}.'/ImageDB/original/EP_Images/'.$TestDBImageName2;
    my $TestFileContent1 = ${GetFileContent($TestDBImage1, { 'binary' => 1 })};
    my $TestFileContent2 = ${GetFileContent($TestDBImage2, { 'binary' => 1 })};
    my $hFiles = {
        'NewFile' => [
            $TestFileContent1,
            $TestFileContent2
         ]
    };

    my $hFileNames = {
        'NewFile' => [
            'animals_4237.jpg',
            'animals_4249.jpg'
        ]
    };
    CreateDirectory($MGPublicPath.'/MyFiles/SomeAddUploadedDir', { 'recursive' => 1 });
    # gallery size before
    $Shop->featureCurrentValue('GallerySize', 0);
    $Shop->featureMaxValue('GallerySize', 1000);
    $MediaGallery->addUploadedFiles('/MyFiles/SomeAddUploadedDir', $hFiles, $hFileNames);
    # check return value - should contain the names of all images

    # files there?
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$TestDBImageName1), "addUploadedFiles: created file 1");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$TestDBImageName2), "addUploadedFiles: created file 2");
    # also the thumbnails?
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$MediaGallery->previewPath($TestDBImageName1, 'mgt')),
        "addUploadedFiles: created file 1 (mgt)");
    ok(ExistsFile($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$MediaGallery->previewPath($TestDBImageName2, 'mgt')),
        "addUploadedFiles: created file 2 (mgt)");
    # check gallery size. new size should equal the sum of the file sizes in the specified directory
    my $TargetDirSize;  # = GetDirectorySize($MGPublicPath.'/MyFiles/SomeAddUploadedDir/');
    $TargetDirSize += GetFileSize($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$TestDBImageName1);
    $TargetDirSize += GetFileSize($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$TestDBImageName2);
    $TargetDirSize += GetFileSize($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$MediaGallery->previewPath($TestDBImageName1, 'mgt'));
    $TargetDirSize += GetFileSize($MGPublicPath.'/MyFiles/SomeAddUploadedDir/'.$MediaGallery->previewPath($TestDBImageName2, 'mgt'));
    # mediagallery uses KBytes instead of bytes, so we can only compare on KB level:
    cmp_ok(int($TargetDirSize/1024), '==', $Shop->featureCurrentValue('GallerySize'), "addUploadedFiles increase gallery size");
}

#========================================================================================
# function     _testImageMethods
# state        private
#----------------------------------------------------------------------------------------
# description  Runs test for image resize and thumbnail creation.
#----------------------------------------------------------------------------------------
# input        $MediaGallery | object | MediaGallery child object of the shop.
#========================================================================================
sub _testImageMethods {
    my ($MediaGallery) = @_;

    my $TestDBImage1 = '/WebRoot/ImageDB/original/EP_Images/electronics_4478.jpg';
    $MediaGallery->copyFileFromImageDb($TestDBImage1);
    cmp_ok($MediaGallery->existsFile('electronics_4478.jpg'), 'eq', 1, "copy file from img db");
    $MediaGallery->deleteFile('electronics_4478.jpg');
    cmp_ok($MediaGallery->existsFile('electronics_4478.jpg'), 'eq', 0, "after delte file is gone");

    my $TestDBImage2 = '/WebRoot/ImageDB/original/EP_Images/botanical_0617.jpg';
    $MediaGallery->copyFileFromImageDb($TestDBImage2, "copytestdir");
    cmp_ok($MediaGallery->existsFile('copytestdir/botanical_0617.jpg'), 'eq', 1, "copy file from img db to subdir");
    $MediaGallery->deleteFile('copytestdir/botanical_0617.jpg');
    cmp_ok($MediaGallery->existsFile('copytestdir/botanical_0617.jpg'), 'eq', 0, "after delte file is gone");

    my $TestDBImage3 = 'http://localhost/WebRoot/ImageDB/thumbnail/EP_Images/construction_4939.jpg';
    $MediaGallery->copyFileFromImageDb($TestDBImage3);
    cmp_ok($MediaGallery->existsFile('construction_4939.jpg'), 'eq', 1, "copy file from img db using url");
    $MediaGallery->deleteFile('clipart_sanata-claus.png');
    cmp_ok($MediaGallery->existsFile('clipart_sanata-claus.png'), 'eq', 0, "after delte file is gone");

    # createThumbnail()
    my $ImgName = 'botanical_02450.jpg';
    my $TestDBImage4 = '/WebRoot/ImageDB/original/EP_Images/'.$ImgName;
    $MediaGallery->copyFileFromImageDb($TestDBImage4);
    $MediaGallery->createThumbnail($ImgName);
    ok($MediaGallery->existsFile($MediaGallery->previewPath($ImgName, 'mgt')), "createThumbnail");

    # deleteThumbnail()
    $MediaGallery->deleteThumbnail($ImgName);
    ok($MediaGallery->existsFile($ImgName), "deleteThumbnail, org image still there");
    ok($MediaGallery->existsFile($MediaGallery->previewPath($ImgName, 'mgt')), "deleteThumbnail, mgt image gone");

    # resizeImage()
    $MediaGallery->resizeImage(
                   $ImgName,
                   $ImgName,
                   100,
                   80,
                   85);
    my $hResizedImageInfo = GetImageInfo($MediaGallery->absolutePath($ImgName));
    # image is quadratic, thus resizing to min(100, 80) gives us 80x80 px (shortest side is size limit as aspect is kept)
    ok($hResizedImageInfo->{'Width'} == 80 && $hResizedImageInfo->{'Height'} == 80, "resizeImage");
}

sub Main {
    ConnectTestSite();
    ok( Init(), "initialization" );

    # Load shop
    my $Shop = LoadObjectByPath($ROOTPATH.'/TestShop1');
    ok( $Shop->id, "create shop " . $Shop->id );

    # Load shop's mediagallery
    my $MediaGallery = $Shop->child('MediaGallery');
    ok( $MediaGallery->id, "fetch media gallery " . $MediaGallery->id );

    _testDirectoryMethods($MediaGallery);
    _testFileMethods($Shop, $MediaGallery);
    _testImageMethods($MediaGallery);
    _testRemainingMethods($Shop, $MediaGallery);

    ok( Cleanup(), "cleanup" );
    DisconnectSite();
}

RunScript( Message => 'long', Sub=> sub {
    RunOnStore( Store => GetTestSite(), Sub => \&Main );
});
